@@ -9,7 +9,7 @@
  * See the file "LICENSE.TERMS" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  *
- * $Id: fcgiapp.h,v 1.4 1999/08/05 21:25:52 roberts Exp $
+ * $Id: fcgiapp.h,v 1.5 2000/11/05 17:09:36 robs Exp $
  */
 
 #ifndef _FCGIAPP_H
@@ -231,6 +231,17 @@ DLLAPI int FCGX_Accept_r(FCGX_Request *request);
  */
 DLLAPI void FCGX_Finish_r(FCGX_Request *request);
 
+/*
+ *----------------------------------------------------------------------
+ *
+ * FCGX_Free --
+ *
+ *      Free the memory and IPC FD associated with the request (multi-thread safe).
+ *
+ *----------------------------------------------------------------------
+ */
+DLLAPI void FCGX_Free(FCGX_Request * request);
+
 /*
  *----------------------------------------------------------------------
  *
